//--------------------------------------------------------------------
// Sistema con CPU, RAM y UART
//--------------------------------------------------------------------

`include "cpuV6.v"
`include "uart.v"

module SYSTEM (
	input clk,		// 25MHz
	input reset,
	
	// UART
	output txd,
	input rxd,
	
	output debug
);

/////////////////////////////// MEMORY ////////////////////////////////

wire [15:0]cdi;	// CPU Data bus, Input
wire [15:0]cdo;	// CPU Data bus, Output
wire [15:0]ca;	// CPU Address bus
wire [15:0]mdo;	// Memory data output


genram #(
`ifdef SIMULATION
	.INITFILE("romGUS.hex"),
`else
	.INITFILE("rand.hex"),
`endif
	.AW(12),	// 4Kword
	.DW(16)
	) ram8 (.clk(~clk), .we( (~iocs) & we), .addr(ca[11:0]),.data_out(mdo),.data_in(cdo));

//////////////////////////////////////////////////
//////					CPU					//////
//////////////////////////////////////////////////


assign cdi = iocs ? iodo : mdo;

wire we;

CPU_GUSV6 GUScpu0( .clk(clk), .we(we), .ca(ca), .cdo(cdo), .cdi(cdi), 
				  .reset(reset), .irq(irq), .ivector(ivector) );

// IRQS
wire [2:0]ivector=3'b001;
wire irq;

// IRQEN reg
reg irqen=0;	
always @(posedge clk or posedge reset ) 
	if (reset) irqen<=0;
	else irqen <= iowe0 ? cdo[0] : irqen;
	
assign irq= uflags[1] & irqen;

/////////////////////////////////////////////////////////
///////////////////// Peripherals ///////////////////////
/////////////////////////////////////////////////////////

wire iocs;		// IO select ($0020 to $003F)

assign iocs=(ca[15:5]==11'b00000000001);
wire iowe0,iowe1,iowe2,iowe3,iowe4,iowe5,iowe6,iowe7;	// Write strobes
assign iowe0=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&(~ca[1])&(~ca[0]);
// assign iowe1=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&(~ca[1])&( ca[0]);
assign iowe2=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&( ca[1])&(~ca[0]);
// assign iowe3=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&( ca[1])&( ca[0]);
// assign iowe4=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&(~ca[1])&(~ca[0]);
// assign iowe5=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&(~ca[1])&( ca[0]);
// assign iowe6=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&( ca[1])&(~ca[0]);
// assign iowe7=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&( ca[1])&( ca[0]);

// write strobes (data written ignored)
// wire clredges = iowe1;	// Clear phase rising edges register (and interrupt)
// wire clrdirty = iowe6;	// Clear DIRTY flag
// Read strobes
wire iore2;		// Clear UART data valid flag
// wire iore5;		// Clear Keyboard KVAL flag
assign iore2=(~we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&( ca[1])&(~ca[0]);
// assign iore5=(~we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&(~ca[1])&( ca[0]);

/////////////////////////////
// I/O data to CPU MUX
reg [15:0]iodo;
always@*
  case (ca[2:0])
     0 : iodo <=  {15'h0,irqen}; // IRQ Enable
     1 : iodo <=  {12'h0,uflags}; // PFLAGS
     2 : iodo <=  { 8'h0,urxdata}; // UART data
     default : iodo <= 16'hxxxx;
  endcase
  
wire [7:0]urxdata;
wire [3:0]uflags;

UART_max_core #(.DIVBITS(8)) uart0 
	( .clk(clk), .d(cdo[7:0]), .wr(iowe2), .rd(iore2), .q(urxdata),
		.rxvalid(uflags[0]), .txrdy(uflags[1]),
		.rxoverr(uflags[2]), .rxframeer(uflags[3]),
		.nstop(1'b0), .paren(1'b0), .parodd(1'b0),
		.bauddiv(8'd217),
		.txd(txd), .rxd(rxd)
	);

endmodule

//----------------------------------------------------------------------------
//-- Generic synchronous RAM memory
//----------------------------------------------------------------------------

module genram (        
    input clk,	// data out on rising edges
    input we,	// Write enable                  
    input wire [AW-1: 0] addr,      
    input wire [DW-1: 0] data_in,
    output reg [DW-1: 0] data_out   
);

parameter INITFILE = "rand.hex";
parameter AW = 13;
parameter DW = 8;

reg [DW-1: 0] mem [0: (1<<AW)-1];

// synchronous read
always @(posedge clk) begin
	data_out <= mem[addr];
end
// write
always @(posedge clk) begin
	if (we) mem[addr]<=data_in;
end

//-- Load contents with hex file
initial begin
  $readmemh(INITFILE, mem);
end

endmodule

